#include "file_transfer.h"
#include "main.h"
#include "resource.h"
#include "cp_time.h"
#include <Commctrl.h>
#include <Shlwapi.h>
#include "log.h"
#include "lists.h"
//-----------------------------------------
static LPTSTR module = TEXT("File Transfer");
//-----------------------------------------------------------------------------------------
static const u32 packet_timeout = 5;
static const u32 connection_timeout = 12000;
static const u32 display_timeout = 600;
//-----------------------------------------------------------------------------------------
enum {
	STATE_WAIT_APPROVE,
	STATE_APPROVING,
	STATE_CANCELING,
	STATE_CANCELED,
	STATE_FINISHED,
	STATE_TIMEOUTED,
	STATE_TRANSFERING,
	STATE_RECEIVING
};
//-----------------------------------------------------------------------------------------
enum {
	TRANSFER_REQUEST,
	TRANSFER_WAIT,
	TRANSFER_ACCEPT,
	TRANSFER_REJECT,
	TRANSFER_CHUNK,
	TRANSFER_CONFIRMATION,
	TRANSFER_CANCEL,
	TRANSFER_CANCELED,
	TRANSFER_FINISH
};
//-----------------------------------------------------------------------------------------
void file_transfer_cancel(transfer_t* tr)
{
	tr->state = STATE_CANCELING;
	tr->timeout = display_timeout;
	tr->chunk_timeout[0] = 10 * packet_timeout;
	LogMessage(module, TEXT("File transfer is cancelled"));
}
//-----------------------------------------------------------------------------------------
void send_command(const NODE* nd, u32 session, u8 command)
{
	NETWORK* net = getNet();
	PACKET_FOR_SEND pack;
	u32 data = (session << 8) | command;

	pack.nd = nd;
	pack.protocol = NETWORK_PROT_FILE_TRANSFER;
	pack.buffer = &data;
	pack.buf_len = sizeof(u32);
	if (net) net->send_packet(&pack);
}
//-----------------------------------------------------------------------------------------
typedef struct {
	transfer_t* tr;
	u64 bytes_sent;
	u64 bytes_total;
} transfer_progress_t;
//----------------------------------------------------------------------------
const u32 scale_size = 5;
const u64 scale_value[scale_size] = { 1, 1000, 1000000, 1000000000, 1000000000000 };
const LPTSTR scale_name[scale_size] = { TEXT("B"), TEXT("KB"), TEXT("MB"), TEXT("GB"), TEXT("TB") };
//----------------------------------------------------------------------------
static u32 find_scale(u64 size)
{
	u32 s;
	for (s = scale_size; s--;)
	{
		if (size >= scale_value[s]) return s;
	}
	return 0;
}
//----------------------------------------------------------------------------
static void fill_progress(TCHAR* str, u32 str_len, transfer_progress_t* tp)
{
	LPTSTR time_str;
	cp_time_t cur_time;
	tp->bytes_sent = tp->tr->chunk_cur*CHUNK_SIZE;

	u32 scale1 = find_scale(tp->bytes_sent);
	u32 scale2 = find_scale(tp->tr->file_size);
	cp_get_time(&cur_time);
	double speed = ((double)tp->bytes_sent * (10000000 / 1000) / (cur_time - tp->tr->start_time)); // speed in KB/sec
	s32 time_left;
	if (speed > 0)
		time_left = (s32)((tp->tr->file_size - tp->bytes_sent) / speed * 1e-3);
	else
		time_left = 24 * 3600;
	if (time_left < 0) time_left = 0;
	if (time_left > 3600)
	{
		time_left /= 3600;
		time_str = TEXT("h");
	}
	else if (time_left > 60)
	{
		time_left /= 60;
		time_str = TEXT("min");
	}
	else
	{
		time_str = TEXT("sec");
	}

	_stprintf_s(str, str_len, TEXT("%s %d %s out of %d %s, speed %.2f KB/sec, time left %d %s"),
		(tp->tr->state == STATE_TRANSFERING) ? TEXT("Transfered") : TEXT("Received"), (u32)(tp->bytes_sent / scale_value[scale1]), scale_name[scale1],
		(u32)(tp->tr->file_size / scale_value[scale2]), scale_name[scale2], speed, time_left, time_str);
}
//----------------------------------------------------------------------------
static void CloseTransfer(HWND hWnd)
{
	transfer_progress_t* tp = (transfer_progress_t*)GetProp(hWnd, TEXT("progress"));
	if ((tp->tr->state != STATE_FINISHED) && (tp->tr->state != STATE_CANCELED) && (tp->tr->state != STATE_TIMEOUTED))
	{
		send_command(&tp->tr->nd, tp->tr->session, (tp->tr->state == STATE_APPROVING) ? TRANSFER_REJECT : TRANSFER_CANCEL);
		if (tp->tr->state == STATE_RECEIVING)
		{
			HANDLE file = tp->tr->file;
			tp->tr->file = NULL;
			CloseHandle(file);
			DeleteFile(tp->tr->file_name);
		}
		tp->tr->state = STATE_CANCELING;
		tp->tr->timeout = display_timeout;
		tp->tr->chunk_timeout[0] = 10*packet_timeout;
		LogMessage(module, TEXT("File transfer is cancelled"));
	}
	else
	{
		tp->tr->timeout = 0;
	}
	tp->tr->hWnd = NULL;
	DestroyWindow(hWnd);
}
//----------------------------------------------------------------------------
static void SaveTransfer(HWND hWnd)
{
	transfer_progress_t* tp = (transfer_progress_t*)GetProp(hWnd, TEXT("progress"));
	OPENFILENAME File;
	TCHAR FileName[MAX_PATH] = TEXT("");

	ZeroMemory(&File, sizeof(OPENFILENAME));
	File.lStructSize = sizeof(OPENFILENAME);
	File.hwndOwner = hWnd;
	File.lpstrFile = FileName;
	File.nMaxFile = MAX_PATH;
	File.lpstrFilter = TEXT("*.*\0*.*\0");
	File.nFilterIndex = 1;
	File.nMaxFileTitle = 0;
	File.Flags = OFN_PATHMUSTEXIST | OFN_OVERWRITEPROMPT;
	getFileTransferPath(FileName);
	cp_strcchcpy(&FileName[cp_strlen(FileName)], MAX_PATH - cp_strlen(FileName), tp->tr->file_name);

	if (!acceptFileTransfer() && !GetOpenFileName(&File))
	{
		CloseTransfer(hWnd);
		return;
	}

	if (acceptFileTransfer())
	{
		int i;
		int n = 1;
		for (i = cp_strlen(tp->tr->file_name); --i;)
			if (tp->tr->file_name[i] == '.') break;
		if (!i) i = cp_strlen(tp->tr->file_name);
		while(PathFileExists(FileName))
		{
			TCHAR num[100];
			cp_snprintf(num,100,TEXT(" (%d)"),n++);
			getFileTransferPath(FileName);
			cp_strncpy(&FileName[cp_strlen(FileName)],tp->tr->file_name,i);
			cp_strcchcpy(&FileName[cp_strlen(FileName)], MAX_PATH - cp_strlen(FileName), num);
			cp_strcchcpy(&FileName[cp_strlen(FileName)], MAX_PATH - cp_strlen(FileName), &tp->tr->file_name[i]);
		}
	}
	tp->tr->file = CreateFile(FileName, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (tp->tr->file == INVALID_HANDLE_VALUE)
	{
		MessageBox(hWnd, TEXT("Can't create file"), TEXT("Error"), MB_ICONERROR | MB_OK);
		return;
	}
	if (tp->tr->file_name) delete[] tp->tr->file_name;
	tp->tr->file_name = new tchar_t[_tcslen(FileName) + 1];
	cp_strcchcpy(tp->tr->file_name, _tcslen(FileName) + 1, FileName);
	tp->tr->timeout = connection_timeout;
	tp->tr->state = STATE_RECEIVING;
	cp_get_time(&tp->tr->start_time);
	send_command(&tp->tr->nd, tp->tr->session, TRANSFER_ACCEPT);
	LogMessage(module, TEXT("File transfer is accepted"));
}
//----------------------------------------------------------------------------
static void Update(HWND hWnd)
{
	TCHAR str[300];
	transfer_progress_t* tp = (transfer_progress_t*)GetProp(hWnd, TEXT("progress"));
	switch (tp->tr->state)
	{
	case STATE_WAIT_APPROVE:
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, TEXT("Waiting"));
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_HIDE);
		break;
	case STATE_APPROVING:
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, TEXT("FILE TRANSFER REQUEST!"));
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_SHOW);
		if (acceptFileTransfer())
		{
			SaveTransfer(hWnd);
		}
		break;
	case STATE_CANCELING:
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, TEXT("Cancelling Transfer"));
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_HIDE);
		break;
	case STATE_CANCELED:
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, TEXT("Transfer is canceled"));
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_HIDE);
		SetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE, GetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE) & ~PBS_MARQUEE);
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 0, 0);
		break;
	case STATE_FINISHED:
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, TEXT("File received"));
		SetDlgItemText(hWnd, IDCANCEL, TEXT("Close"));
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_HIDE);
		SetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE, GetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE) & ~PBS_MARQUEE);
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 100, 0);
		break;
	case STATE_TIMEOUTED:
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, TEXT("Connection timeout"));
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_HIDE);
		SetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE, GetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE) & ~PBS_MARQUEE);
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 0, 0);
		break;
	case STATE_TRANSFERING:
	case STATE_RECEIVING:
		ShowWindow(GetDlgItem(hWnd, IDC_SAVEAS), SW_HIDE);
		fill_progress(str, 300, tp);
		SetDlgItemText(hWnd, IDC_TRANSFERTEXT, str);
		SetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE, GetWindowLong(GetDlgItem(hWnd, IDC_PROGRESS), GWL_STYLE) & ~PBS_MARQUEE);
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, (u32)(tp->tr->chunk_cur * 100 / tp->tr->chunk_total), 0);
		break;
	}
}
//-----------------------------------------------------------------------------------------
static void TransferInit(HWND hWnd, transfer_t* tr)
{
	transfer_progress_t* tp = new transfer_progress_t;
	SetProp(hWnd, TEXT("progress"), tp);
	tp->tr = tr;
	tr->hWnd = hWnd;
	SetWindowText(hWnd, tp->tr->file_name);
	SetTimer(hWnd, 1, 333, NULL);
	Update(hWnd);
	SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETMARQUEE, 1, 0);
}
//----------------------------------------------------------------------------
static void TransferDestroy(HWND hWnd)
{
	transfer_progress_t* tp = (transfer_progress_t*)RemoveProp(hWnd, TEXT("progress"));
	KillTimer(hWnd, 1);
	delete tp;
}
//----------------------------------------------------------------------------
INT_PTR CALLBACK TransferProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG: TransferInit(hWnd, (transfer_t*)lParam); return TRUE;
	case WM_DESTROY: TransferDestroy(hWnd); return 0;
	case WM_CLOSE: CloseTransfer(hWnd); return TRUE;
	case WM_TIMER: Update(hWnd); return TRUE;
	case WM_UPDATE_PROGRESS: Update(hWnd); return TRUE;
	case WM_CLOSEWINDOW: DestroyWindow(hWnd); return TRUE;
	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_SAVEAS:
			SaveTransfer(hWnd);
			return TRUE;
		case IDCANCEL:
			CloseTransfer(hWnd);
			return TRUE;
		}
		return FALSE;
	}
	return FALSE;
}
//-----------------------------------------------------------------------------------------
void free_transfer(transfer_t* tr)
{
	if (tr->file != INVALID_HANDLE_VALUE) CloseHandle(tr->file);
	if (tr->hWnd) PostMessage(tr->hWnd, WM_CLOSEWINDOW, 0, 0);
	if (tr->file_name) delete[] tr->file_name;
	delete tr;
}
//-----------------------------------------------------------------------------------------
FILE_TRANSFER::FILE_TRANSFER(HWND mWnd)
{
	mainWnd = mWnd;
	transfers = NULL;
	InitializeCriticalSection(&cs);
	stop = CreateEvent(NULL, TRUE, FALSE, NULL);
	timer = CreateThread(NULL, 0, timeout_trhead, this, 0, NULL);
}
//-----------------------------------------------------------------------------------------
FILE_TRANSFER::~FILE_TRANSFER()
{
	transfer_t *tr1, *tr2;
	SetEvent(stop);
	WaitForSingleObject(timer, INFINITE);
	EnterCriticalSection(&cs);
	for (tr1 = tr2 = transfers; tr1; tr1 = tr2)
	{
		tr2 = tr1->next;
		free_transfer(tr1);
	}
	transfers = NULL;
	LeaveCriticalSection(&cs);
	DeleteCriticalSection(&cs);
}
//-----------------------------------------------------------------------------------------
transfer_t* FILE_TRANSFER::new_transfer(void)
{
	transfer_t* tr;
	if (!transfers)
	{
		tr = transfers = new transfer_t;
	}
	else
	{
		for (tr = transfers; tr->next; tr = tr->next);
		tr->next = new transfer_t;
		tr = tr->next;
	}
	if (tr)
	{
		tr->next = NULL;
		tr->file = INVALID_HANDLE_VALUE;
		tr->hWnd = NULL;
		tr->file_name = NULL;
	}
	return tr;
}
//-----------------------------------------------------------------------------------------
void FILE_TRANSFER::transfer(const NODE* nd, pctstr_t file_name)
{
	transfer_t* tr;
	int k;
	LARGE_INTEGER file_size;
	HANDLE file = CreateFile(file_name, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (file == INVALID_HANDLE_VALUE)
	{
		MessageBox(mainWnd, TEXT("Can't open file for transfer"), TEXT("Error"), MB_OK | MB_ICONERROR);
		return;
	}
	GetFileSizeEx(file, &file_size);
	if (!file_size.QuadPart)
	{
		CloseHandle(file);
		MessageBox(mainWnd, TEXT("File is empty"), TEXT("Error"), MB_OK | MB_ICONERROR);
		return;
	}
	EnterCriticalSection(&cs);
	tr = new_transfer();
	if (!tr)
	{
		LeaveCriticalSection(&cs);
		CloseHandle(file);
		MessageBox(mainWnd, TEXT("Don't have enough memory for new transfer"), TEXT("Error"), MB_OK | MB_ICONERROR);
		return;
	}
	tr->file = file;
	tr->state = STATE_WAIT_APPROVE;
	cheap_random_number((u8*)&tr->session, sizeof(tr->session));
	tr->session &= 0xFFFFFF; // only 3 bytes;
	tr->nd = *nd;
	for (k = 0; k < CHUNK_NUM; k++)
	{
		tr->chunk_seq[k] = 0xFFFFFFFFFFFFFFFF;
	}
	tr->file_size = file_size.QuadPart;
	tr->chunk_total = (tr->file_size + CHUNK_SIZE - 1) / CHUNK_SIZE;
	tr->chunk_cur = 0;
	tr->timeout = connection_timeout;

	for (k = (int)_tcslen(file_name) - 1; k > 0; k--)
	{
		if (file_name[k] == '\\')
		{
			k++;
			break;
		}
	}
	tr->file_name = new tchar_t[_tcslen(file_name) - k + 1];
	cp_strcchcpy(tr->file_name, _tcslen(file_name) - k + 1, &file_name[k]);

	tr->chunk[0][0] = TRANSFER_REQUEST | (tr->session << 8);
	tr->chunk[0][1] = (u32)(tr->file_size >> 32);
	tr->chunk[0][2] = (u32)(tr->file_size & 0xFFFFFFFF);
	tr->chunk_size[0] = 3 * sizeof(u32)+((u32)_tcslen(file_name) - k)*sizeof(tchar_t);
	if (tr->chunk_size[0] > NETWORK_SIZE_USEFUL)
		tr->chunk_size[0] = NETWORK_SIZE_USEFUL;
	memcpy(&tr->chunk[0][3], &file_name[k], tr->chunk_size[0] - 3 * sizeof(u32));
	tr->chunk_timeout[0] = 10 * packet_timeout;
	LeaveCriticalSection(&cs);
	PostMessage(mainWnd, WM_FILETRANSFER, 0, (LPARAM)tr);

	NETWORK* net = getNet();
	PACKET_FOR_SEND pack;
	pack.buffer = tr->chunk[0];
	pack.buf_len = tr->chunk_size[0];
	pack.nd = nd;
	pack.protocol = NETWORK_PROT_FILE_TRANSFER;
	if (net) net->send_packet(&pack);
	LogMessage(module, TEXT("File transfer is started"));
}
//-----------------------------------------------------------------------------------------
void send_available_packets(transfer_t* tr)
{
	int i;
	PACKET_FOR_SEND pack;
	NETWORK* net = getNet();
	pack.nd = &tr->nd;
	pack.protocol = NETWORK_PROT_FILE_TRANSFER;
	for (i = 0; i < CHUNK_NUM; i++)
	{
		if ((tr->chunk_seq[i] != 0xFFFFFFFFFFFFFFFF) && !tr->chunk_timeout[i])
		{
			pack.buffer = tr->chunk[i];
			pack.buf_len = tr->chunk_size[i];
			if (net) net->send_packet(&pack);
			tr->chunk_timeout[i] = packet_timeout;
		}
	}
}
//-----------------------------------------------------------------------------------------
void send_portion(transfer_t* tr)
{
	int i;
	for (i = 0; (i < CHUNK_NUM) && (tr->chunk_cur < tr->chunk_total); i++)
	{
		if (tr->chunk_seq[i] == 0xFFFFFFFFFFFFFFFF)
		{
			u32 bytes = (tr->chunk_cur == tr->chunk_total - 1) ? (u32)(tr->file_size - tr->chunk_cur*CHUNK_SIZE) : CHUNK_SIZE;
			DWORD read_bytes = 0;
			tr->chunk_seq[i] = tr->chunk_cur;
			tr->chunk[i][0] = TRANSFER_CHUNK | (tr->session << 8);
			tr->chunk[i][1] = (u32)(tr->chunk_cur >> 32);
			tr->chunk[i][2] = (u32)(tr->chunk_cur & 0xFFFFFFFF);
			ReadFile(tr->file, &tr->chunk[i][3], bytes, &read_bytes, NULL);
			tr->chunk_size[i] = read_bytes + 3 * sizeof(u32);
			tr->chunk_timeout[i] = 0; // request to send right now
			tr->chunk_cur++;
		}
	}
	send_available_packets(tr);
	// check that transfer is finished
	if (tr->chunk_cur >= tr->chunk_total)
	{
		bool finished = true;
		for (i = 0; i < CHUNK_NUM; i++)
		{
			if (tr->chunk_seq[i] != 0xFFFFFFFFFFFFFFFF)
			{
				finished = false;
				break;
			}
		}
		if (finished)
		{
			tr->state = STATE_FINISHED;
			tr->timeout = display_timeout;
			CloseHandle(tr->file);
			tr->file = NULL;
			PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
			LogMessage(module, TEXT("File transfer is finished"));
		}
	}
}
//-----------------------------------------------------------------------------------------
void send_confirmation(transfer_t* tr, u64 chunk_num)
{
	NETWORK* net = getNet();
	PACKET_FOR_SEND pack;
	u32 data[3];
	pack.nd = &tr->nd;
	pack.protocol = NETWORK_PROT_FILE_TRANSFER;
	pack.buffer = data;
	pack.buf_len = sizeof(data);
	data[0] = (tr->session << 8) | TRANSFER_CONFIRMATION;
	data[1] = (u32)(chunk_num >> 32);
	data[2] = (u32)(chunk_num & 0xFFFFFFFF);
	if (net) net->send_packet(&pack);
}
//-----------------------------------------------------------------------------------------
void receive_portion(transfer_t* tr, const u32* packet, u32 size)
{
	u64 chunk_num = ((u64)packet[1] << 32) | packet[2];
	int i,j;
	DWORD wrb;

	if (chunk_num < tr->chunk_cur) // we've already received that packet
	{
		send_confirmation(tr, chunk_num);
	}
	else if (chunk_num == tr->chunk_cur) // write to file immidiately and check other buffers
	{
		WriteFile(tr->file, &packet[3], size - 3 * sizeof(u32), &wrb, NULL);
		send_confirmation(tr, chunk_num);
		tr->chunk_cur++;
		for (j = 0; j < CHUNK_NUM; j++)
		{
			for (i = 0; i < CHUNK_NUM; i++) // twice because of random order
			{
				if (tr->chunk_seq[i] == tr->chunk_cur)
				{
					WriteFile(tr->file, &tr->chunk[i][3], tr->chunk_size[i] - 3 * sizeof(u32), &wrb, NULL);
					tr->chunk_cur++;
					tr->chunk_seq[i] = 0xFFFFFFFFFFFFFFFF;
				}
			}
		}
	}
	else // put in list for future processing
	{
		for (i = 0; i < CHUNK_NUM; i++)
		{
			if (tr->chunk_seq[i] == 0xFFFFFFFFFFFFFFFF)
			{
				CopyMemory(tr->chunk[i], packet, size);
				tr->chunk_size[i] = size;
				tr->chunk_seq[i] = chunk_num;
				send_confirmation(tr, chunk_num);
				break;
			}
		}
	}
}
//-----------------------------------------------------------------------------------------
void FILE_TRANSFER::parse_packet(const u32* packet, u32 size, const NODE* nd, int acceptAll)
{
	u8 cmd;
	u32 session;
	transfer_t* tr;
	int i;

	if (size < sizeof(u32)) return;
	cmd = (u8)(packet[0] & 0xFF);
	session = packet[0]>>8;
	EnterCriticalSection(&cs);
	for (tr = transfers; tr; tr = tr->next)
	{
		if (isIdIdentical(nd->id, tr->nd.id) && (session == tr->session))
			break;
	}
	switch (cmd)
	{
	case TRANSFER_REQUEST:
		if (tr) // already received such request
		{
			switch (tr->state)
			{
			case STATE_APPROVING:
				send_command(nd, session, TRANSFER_WAIT);
				break;
			case STATE_RECEIVING:
				send_command(nd, session, TRANSFER_ACCEPT);
				break;
			default:
				send_command(nd, session, TRANSFER_REJECT);
				break;
			}
		}
		else if ((size >= 3 * sizeof(u32)) && (acceptAll || isInLikes(nd))) // create new transfer
		{
			tr = new_transfer();
			if (tr)
			{
				tr->session = session;
				tr->nd = *nd;
				tr->state = STATE_APPROVING;
				tr->timeout = display_timeout;
				tr->chunk_cur = 0;
				tr->file_size = ((u64)packet[1] << 32) | packet[2];
				tr->chunk_total = (tr->file_size + CHUNK_SIZE - 1) / CHUNK_SIZE;
				for (i = 0; i < CHUNK_NUM; i++)
				{
					tr->chunk_seq[i] = 0xFFFFFFFFFFFFFFFF;
				}
				tr->file_name = new tchar_t[(size - 3 * sizeof(u32)) / sizeof(tchar_t) + 1];
				CopyMemory(tr->file_name, &packet[3], size - 3 * sizeof(u32));
				tr->file_name[(size - 3 * sizeof(u32)) / sizeof(tchar_t)] = 0;
				PostMessage(mainWnd, WM_FILETRANSFER, 0, (LPARAM)tr);
				LogMessage(module, TEXT("Received request for file transfer"));
			}
		}
		else // any other condition leads to transaction cancellation
		{
			send_command(nd, session, TRANSFER_CANCEL);
		}
		break;
	case TRANSFER_WAIT:
		if (tr && (tr->state == STATE_WAIT_APPROVE))
		{
			tr->timeout = connection_timeout;
		}
		break;
	case TRANSFER_ACCEPT:
		if (tr && (tr->state == STATE_WAIT_APPROVE))
		{
			tr->state = STATE_TRANSFERING;
			tr->timeout = connection_timeout;
			cp_get_time(&tr->start_time);
			send_portion(tr);
			PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
			LogMessage(module, TEXT("File transfer was accepted"));
		}
		break;
	case TRANSFER_REJECT:
		if (tr && (tr->state != STATE_CANCELED))
		{
			tr->state = STATE_CANCELED;
			tr->timeout = display_timeout;
			PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
			LogMessage(module, TEXT("File transfer was rejected"));
		}
		break;
	case TRANSFER_CHUNK:
		if (tr && (tr->state == STATE_RECEIVING) && (size >= 3*sizeof(u32)))
		{
			receive_portion(tr, packet, size);
			if (tr->chunk_cur >= tr->chunk_total) // finished
			{
				tr->state = STATE_FINISHED;
				tr->timeout = display_timeout;
				CloseHandle(tr->file);
				tr->file = NULL;
				PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
				LogMessage(module, TEXT("File transfer is finished"));
				TCHAR user_info[user_info_max_length_bytes / sizeof(TCHAR)];
				getNet()->get_user_info(&tr->nd, user_info);
				getUserList()->append_download(&tr->nd,tr->file_name,user_info);
				PostMessage(g_DownloadsList,WM_UPDATE_PROGRESS, 0, 0);
			}
			else
			{
				tr->timeout = connection_timeout;
			}
		}
		else if (tr && (tr->state == STATE_FINISHED))
		{
			send_command(nd, session, TRANSFER_FINISH);
		}
		else // any other condition leads to transaction cancellation
		{
			send_command(nd, session, TRANSFER_CANCEL);
		}
		break;
	case TRANSFER_CONFIRMATION:
		if (tr && (tr->state == STATE_TRANSFERING))
		{
			if (size >= 3 * sizeof(u32))
			{
				u64 chunk_num = ((u64)packet[1] << 32) | packet[2];
				for (i = 0; i < CHUNK_NUM; i++)
				{
					if (tr->chunk_seq[i] == chunk_num)
					{
						tr->chunk_seq[i] = 0xFFFFFFFFFFFFFFFF; // packet receprion was confirmed
						break;
					}
				}
			}
			send_portion(tr);
		}
		else // any other condition leads to transaction cancellation
		{
			send_command(nd, session, TRANSFER_CANCEL);
		}
		break;
	case TRANSFER_FINISH:
		if (tr && ((tr->state == STATE_TRANSFERING) || (tr->state == STATE_RECEIVING)))
		{
			tr->state = STATE_FINISHED;
			tr->timeout = display_timeout;
			PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
			LogMessage(module, TEXT("File transfer was finished"));
		}
		break;
	case TRANSFER_CANCEL:
		if (tr && (tr->state != STATE_CANCELED))
		{
			tr->state = STATE_CANCELED;
			tr->timeout = display_timeout;
			PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
			LogMessage(module, TEXT("File transfer was cancelled"));
		}
		send_command(nd, session, TRANSFER_CANCELED);
		break;
	case TRANSFER_CANCELED:
		if (tr && (tr->state == STATE_CANCELING))
		{
			tr->state = STATE_CANCELED;
			tr->timeout = display_timeout;
			PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
		}
		break;
	}
	LeaveCriticalSection(&cs);
}
//-----------------------------------------------------------------------------------------
DWORD CALLBACK timeout_trhead(void* foo)
{
	FILE_TRANSFER* ft = (FILE_TRANSFER*)foo;
	transfer_t* tr;
	PACKET_FOR_SEND pack;
	NETWORK* net;
	int i;

	while (WAIT_TIMEOUT == WaitForSingleObject(ft->stop, 50))
	{
		EnterCriticalSection(&ft->cs);
		for (tr = ft->transfers; tr;)
		{
			switch (tr->state)
			{
			case STATE_WAIT_APPROVE:
				if (!tr->timeout) // wait approval timeout
				{
					tr->state = STATE_TIMEOUTED;
					tr->timeout = display_timeout;
					PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
					LogMessage(module, TEXT("File transfer timed out"));
				}
				else
				{
					tr->timeout--;
					if (!tr->chunk_timeout[0])
					{
						net = getNet();
						pack.buffer = tr->chunk[0];
						pack.buf_len = tr->chunk_size[0];
						pack.nd = &tr->nd;
						pack.protocol = NETWORK_PROT_FILE_TRANSFER;
						if (net) net->send_packet(&pack);
						tr->chunk_timeout[0] = 10 * packet_timeout;
					}
					else
					{
						tr->chunk_timeout[0]--;
					}
				}
				break;
			case STATE_TRANSFERING:
				if (!tr->timeout) // wait connection timeout
				{
					tr->state = STATE_TIMEOUTED;
					tr->timeout = display_timeout;
					PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
					LogMessage(module, TEXT("File transfer timed out"));
				}
				else
				{
					tr->timeout--;
					for (i = 0; i < CHUNK_NUM; i++)
					{
						if ((tr->chunk_seq[i] != 0xFFFFFFFFFFFFFFFF) && tr->chunk_timeout[i])
						{
							tr->chunk_timeout[i]--;
						}
					}
					send_available_packets(tr);
				}
				break;
			case STATE_APPROVING:
				if (!tr->timeout) // wait approval timeout
				{
					tr->state = STATE_TIMEOUTED;
					tr->timeout = display_timeout;
					PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
					LogMessage(module, TEXT("File transfer timed out"));
				}
				else
				{
					tr->timeout--;
				}
				break;
			case STATE_CANCELING:
				if (!tr->timeout) // wait connection timeout
				{
					tr->state = STATE_CANCELED;
					tr->timeout = display_timeout;
					PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
				}
				else
				{
					tr->timeout--;
					if (!tr->chunk_timeout[0])
					{
						send_command(&tr->nd, tr->session, TRANSFER_CANCEL);
						tr->chunk_timeout[0] = 10*packet_timeout;
					}
					else
					{
						tr->chunk_timeout[0]--;
					}
				}
				break;
			case STATE_RECEIVING:
				if (!tr->timeout) // wait connection timeout
				{
					CloseHandle(tr->file);
					tr->file = NULL;
					DeleteFile(tr->file_name);
					tr->state = STATE_TIMEOUTED;
					tr->timeout = display_timeout;
					PostMessage(tr->hWnd, WM_UPDATE_PROGRESS, 0, 0);
					LogMessage(module, TEXT("File transfer timed out"));
				}
				else
				{
					tr->timeout--;
				}
				break;
			case STATE_FINISHED:
			case STATE_CANCELED:
			case STATE_TIMEOUTED:
				if (!tr->timeout) // display timeout
				{
					transfer_t* dead_tr = tr; // we will kill this transfer later
					if (ft->transfers == tr)
					{
						ft->transfers = tr = tr->next;
					}
					else
					{
						for (transfer_t* pred = ft->transfers; pred->next; pred = pred->next)
						{
							if (pred->next == tr)
							{
								pred->next = tr = tr->next;
								break;
							}
						}
					}
					free_transfer(dead_tr);
					LogMessage(module, TEXT("File transfer closed"));
					continue;
				}
				else
				{
					tr->timeout--;
				}
				break;
			}
			tr = tr->next; // not in for() because some states will destroy items
		}
		LeaveCriticalSection(&ft->cs);
	}
	return 0;
}
//-----------------------------------------------------------------------------------------
void FileTransfer(HWND hWnd, NODE *nd)
{
	FILE_TRANSFER* ft = getFileTransfer();
	TCHAR FileName[MAX_PATH] = TEXT("");
	OPENFILENAME File;
	ZeroMemory(&File, sizeof(OPENFILENAME));
	File.lStructSize = sizeof(OPENFILENAME);
	File.hwndOwner = hWnd;
	File.lpstrFile = FileName;
	File.nMaxFile = MAX_PATH;
	File.lpstrFilter = TEXT("*.*\0*.*\0");
	File.nFilterIndex = 1;
	File.nMaxFileTitle = 0;
	File.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

	if (!GetOpenFileName(&File))
		return;
	ft->transfer(nd, FileName);
}
//-----------------------------------------------------------------------------------------
